@@ -24,13 +24,12 @@
 
 """
 
-from copy import copy
+import numpy
 
-from pyclustering.cluster import cluster_visualizer
+from copy import copy
 
 from pyclustering.utils import euclidean_distance_square
 from pyclustering.utils import manhattan_distance
-from pyclustering.utils import list_math_addition, list_math_subtraction, list_math_multiplication
 from pyclustering.utils import linear_sum, square_sum
 
 from enum import IntEnum
@@ -87,7 +86,7 @@ class cfentry:
     @see cftree
     
     """
-    
+
     @property
     def number_points(self):
         """!
@@ -97,7 +96,7 @@ class cfentry:
         
         """
         return self.__number_points
-    
+
     @property
     def linear_sum(self):
         """!
@@ -108,7 +107,7 @@ class cfentry:
         """
         
         return self.__linear_sum
-    
+
     @property
     def square_sum(self):
         """!
@@ -118,8 +117,8 @@ class cfentry:
         
         """
         return self.__square_sum
-    
-    
+
+
     def __init__(self, number_points, linear_sum, square_sum):
         """!
         @brief CF-entry constructor.
@@ -131,38 +130,39 @@ class cfentry:
         """
         
         self.__number_points = number_points
-        self.__linear_sum = linear_sum
+        self.__linear_sum = numpy.array(linear_sum)
         self.__square_sum = square_sum
         
         self.__centroid = None
         self.__radius = None
         self.__diameter = None
-    
-    
+
+
     def __copy__(self):
         """!
-        @returns (cfentry) Makes copy of the cfentry instance.
+        @returns (cfentry) Makes copy of the CF-entry instance.
         
         """
         return cfentry(self.__number_points, self.__linear_sum, self.__square_sum)
-    
-    
+
+
     def __repr__(self):
         """!
-        @return (string) Default cfentry representation.
+        @return (string) Returns CF-entry representation.
         
         """
-        return 'CF (%s, %s, %0.2f) [%s]' % (self.number_points, self.linear_sum, self.__square_sum, hex(id(self)))
-    
-    
+        return "CF-Entry (N: '%s', LS: '%s', D: '%s')" % \
+               (self.number_points, self.linear_sum, str(self.get_diameter()))
+
+
     def __str__(self):
         """!
         @brief Default cfentry string representation.
         
         """
         return self.__repr__()
-    
-    
+
+
     def __add__(self, entry):
         """!
         @brief Overloaded operator add. Performs addition of two clustering features.
@@ -174,33 +174,12 @@ class cfentry:
         """
         
         number_points = self.number_points + entry.number_points
-        result_linear_sum = list_math_addition(self.linear_sum, entry.linear_sum)
+        result_linear_sum = numpy.add(self.linear_sum, entry.linear_sum)
         result_square_sum = self.square_sum + entry.square_sum
         
         return cfentry(number_points, result_linear_sum, result_square_sum)
-    
-    
-    def __sub__(self, entry):
-        """!
-        @brief Overloaded operator sub. Performs substraction of two clustering features.
-        @details Substraction can't be performed with clustering feature whose description is less then substractor.
-        
-        @param[in] entry (cfentry): Entry that is substracted from the current.
-        
-        @return (cfentry) Result of substraction of two clustering features.
-        
-        """
-                
-        number_points = self.number_points - entry.number_points
-        result_linear_sum = list_math_subtraction(self.linear_sum, entry.linear_sum)
-        result_square_sum = self.square_sum - entry.square_sum
-        
-        if (number_points < 0) or (result_square_sum < 0):
-            raise NameError("Substruction with negative result is not possible for clustering features.")
-        
-        return cfentry(number_points, result_linear_sum, result_square_sum)
-    
-    
+
+
     def __eq__(self, entry):
         """!
         @brief Overloaded operator eq. 
@@ -221,8 +200,8 @@ class cfentry:
             result &= ((self.linear_sum[index_dimension] + tolerance > entry.linear_sum[index_dimension]) and (self.linear_sum[index_dimension] - tolerance < entry.linear_sum[index_dimension]))
         
         return result
-    
-    
+
+
     def get_distance(self, entry, type_measurement):
         """!
         @brief Calculates distance between two clusters in line with measurement type.
@@ -255,24 +234,21 @@ class cfentry:
 
         else:
             raise ValueError("Unsupported type of measurement '%s' is specified." % type_measurement)
-    
-        
+
+
     def get_centroid(self):
         """!
         @brief Calculates centroid of cluster that is represented by the entry. 
         @details It's calculated once when it's requested after the last changes.
         
-        @return (list) Centroid of cluster that is represented by the entry.
+        @return (array_like) Centroid of cluster that is represented by the entry.
         
         """
         
         if self.__centroid is not None:
             return self.__centroid
-        
-        self.__centroid = [0] * len(self.linear_sum)
-        for index_dimension in range(0, len(self.linear_sum)):
-            self.__centroid[index_dimension] = self.linear_sum[index_dimension] / self.number_points
-        
+
+        self.__centroid = numpy.divide(self.linear_sum, self.number_points)
         return self.__centroid
     
     
@@ -287,22 +263,18 @@ class cfentry:
         
         if self.__radius is not None:
             return self.__radius
-        
+
+        N = self.number_points
         centroid = self.get_centroid()
         
         radius_part_1 = self.square_sum
+        radius_part_2 = 2.0 * numpy.dot(self.linear_sum, centroid)
+        radius_part_3 = N * numpy.dot(centroid, centroid)
         
-        if type(centroid) == list:
-            radius_part_2 = 2.0 * sum(list_math_multiplication(self.linear_sum, centroid))
-            radius_part_3 = self.number_points * sum(list_math_multiplication(centroid, centroid))
-        else:
-            radius_part_2 = 2.0 * self.linear_sum * centroid
-            radius_part_3 = self.number_points * centroid * centroid
-        
-        self.__radius = ((1.0 / self.number_points) * (radius_part_1 - radius_part_2 + radius_part_3)) ** 0.5
+        self.__radius = ((1.0 / N) * (radius_part_1 - radius_part_2 + radius_part_3)) ** 0.5
         return self.__radius
-        
-    
+
+
     def get_diameter(self):
         """!
         @brief Calculates diameter of cluster that is represented by the entry.
@@ -315,15 +287,16 @@ class cfentry:
         if self.__diameter is not None:
             return self.__diameter
 
-        if type(self.linear_sum) == list:
-            diameter_part = self.square_sum * self.number_points - 2.0 * sum(list_math_multiplication(self.linear_sum, self.linear_sum)) + self.square_sum * self.number_points
+        diameter_part = self.square_sum * self.number_points - 2.0 * numpy.dot(self.linear_sum, self.linear_sum) + self.square_sum * self.number_points
+
+        if diameter_part < 0.000000001:
+            self.__diameter = 0.0
         else:
-            diameter_part = self.square_sum * self.number_points - 2.0 * self.linear_sum * self.linear_sum + self.square_sum * self.number_points
-            
-        self.__diameter = (diameter_part / (self.number_points * (self.number_points - 1))) ** 0.5
+            self.__diameter = (diameter_part / (self.number_points * (self.number_points - 1))) ** 0.5
+
         return self.__diameter
-    
-        
+
+
     def __get_average_inter_cluster_distance(self, entry):
         """!
         @brief Calculates average inter cluster distance between current and specified clusters.
@@ -334,11 +307,11 @@ class cfentry:
         
         """
         
-        linear_part_distance = sum(list_math_multiplication(self.linear_sum, entry.linear_sum))
-        
+        linear_part_distance = numpy.dot(self.linear_sum, entry.linear_sum)
+
         return ((entry.number_points * self.square_sum - 2.0 * linear_part_distance + self.number_points * entry.square_sum) / (self.number_points * entry.number_points)) ** 0.5
-    
-    
+
+
     def __get_average_intra_cluster_distance(self, entry):
         """!
         @brief Calculates average intra cluster distance between current and specified clusters.
@@ -349,10 +322,10 @@ class cfentry:
         
         """
         
-        linear_part_first = list_math_addition(self.linear_sum, entry.linear_sum)
+        linear_part_first = numpy.add(self.linear_sum, entry.linear_sum)
         linear_part_second = linear_part_first
         
-        linear_part_distance = sum(list_math_multiplication(linear_part_first, linear_part_second))
+        linear_part_distance = numpy.dot(linear_part_first, linear_part_second)
         
         general_part_distance = 2.0 * (self.number_points + entry.number_points) * (self.square_sum + entry.square_sum) - 2.0 * linear_part_distance
         
@@ -369,15 +342,15 @@ class cfentry:
         
         """
                 
-        linear_part_12 = list_math_addition(self.linear_sum, entry.linear_sum)
+        linear_part_12 = numpy.add(self.linear_sum, entry.linear_sum)
         variance_part_first = (self.square_sum + entry.square_sum) - \
-            2.0 * sum(list_math_multiplication(linear_part_12, linear_part_12)) / (self.number_points + entry.number_points) + \
-            (self.number_points + entry.number_points) * sum(list_math_multiplication(linear_part_12, linear_part_12)) / (self.number_points + entry.number_points)**2.0
+            2.0 * numpy.dot(linear_part_12, linear_part_12) / (self.number_points + entry.number_points) + \
+            (self.number_points + entry.number_points) * numpy.dot(linear_part_12, linear_part_12) / (self.number_points + entry.number_points)**2.0
         
-        linear_part_11 = sum(list_math_multiplication(self.linear_sum, self.linear_sum))
+        linear_part_11 = numpy.dot(self.linear_sum, self.linear_sum)
         variance_part_second = -(self.square_sum - (2.0 * linear_part_11 / self.number_points) + (linear_part_11 / self.number_points))
         
-        linear_part_22 = sum(list_math_multiplication(entry.linear_sum, entry.linear_sum))
+        linear_part_22 = numpy.dot(entry.linear_sum, entry.linear_sum)
         variance_part_third = -(entry.square_sum - (2.0 / entry.number_points) * linear_part_22 + entry.number_points * (1.0 / entry.number_points ** 2.0) * linear_part_22)
 
         return variance_part_first + variance_part_second + variance_part_third
@@ -389,13 +362,12 @@ class cfnode:
     
     """
     
-    def __init__(self, feature, parent, payload):
+    def __init__(self, feature, parent):
         """!
         @brief Constructor of abstract CF node.
         
         @param[in] feature (cfentry): Clustering feature of the created node.
         @param[in] parent (cfnode): Parent of the created node.
-        @param[in] payload (*): Data that is stored by the node.
         
         """
         
@@ -407,9 +379,6 @@ class cfnode:
         
         ## Type node (leaf or non-leaf).
         self.type = cfnode_type.CFNODE_DUMMY
-        
-        ## Payload of node where user data can be stored.
-        self.payload = payload
     
     
     def __repr__(self):
@@ -458,18 +427,17 @@ class non_leaf_node(cfnode):
         return self.__successors
     
     
-    def __init__(self, feature, parent, successors, payload):
+    def __init__(self, feature, parent, successors):
         """!
         @brief Create CF Non-leaf node.
         
         @param[in] feature (cfentry): Clustering feature of the created node.
         @param[in] parent (non_leaf_node): Parent of the created node.
         @param[in] successors (list): List of successors of the node.
-        @param[in] payload (*): Data that is stored by the node.
         
         """
                 
-        super().__init__(feature, parent, payload)
+        super().__init__(feature, parent)
         
         ## Node type in CF tree that is CFNODE_NONLEAF for non leaf node.
         self.type = cfnode_type.CFNODE_NONLEAF
@@ -505,22 +473,8 @@ class non_leaf_node(cfnode):
         self.successors.append(successor)
         
         successor.parent = self
-    
-    
-    def remove_successor(self, successor):
-        """!
-        @brief Remove successor from the node.
-        
-        @param[in] successor (cfnode): Successor for removing.
-        
-        """
-        
-        self.feature -= successor.feature
-        self.successors.append(successor)
-        
-        successor.parent = self
-    
-    
+
+
     def merge(self, node):
         """!
         @brief Merge non-leaf node to the current.
@@ -562,7 +516,7 @@ class non_leaf_node(cfnode):
                     farthest_node1 = candidate1
                     farthest_node2 = candidate2
         
-                    return [farthest_node1, farthest_node2]
+        return [farthest_node1, farthest_node2]
     
     
     def get_nearest_successors(self, type_measurement):
@@ -609,18 +563,17 @@ class leaf_node(cfnode):
         return self.__entries
     
     
-    def __init__(self, feature, parent, entries, payload):
+    def __init__(self, feature, parent, entries):
         """!
         @brief Create CF Leaf node.
         
         @param[in] feature (cfentry): Clustering feature of the created node.
         @param[in] parent (non_leaf_node): Parent of the created node.
         @param[in] entries (list): List of entries of the node.
-        @param[in] payload (*): Data that is stored by the node.
         
         """
         
-        super().__init__(feature, parent, payload)
+        super().__init__(feature, parent)
         
         ## Node type in CF tree that is CFNODE_LEAF for leaf node.
         self.type = cfnode_type.CFNODE_LEAF
@@ -637,7 +590,8 @@ class leaf_node(cfnode):
         for entry in self.entries:
             text_entries += "\t" + str(entry) + "\n"
         
-        return 'Leaf-node %s, parent %s, feature %s, entries: %d %s' % (hex(id(self)), self.parent, self.feature, len(self.entries), text_entries)
+        return "Leaf-node: '%s', parent: '%s', feature: '%s', entries: '%d'" % \
+               (str(hex(id(self))), self.parent, self.feature, len(self.entries))
     
     
     def __str__(self):
@@ -658,20 +612,8 @@ class leaf_node(cfnode):
 
         self.feature += entry
         self.entries.append(entry)
-        
-    
-    def remove_entry(self, entry):
-        """!
-        @brief Remove clustering feature from the leaf node.
-        
-        @param[in] entry (cfentry): Clustering feature.
-        
-        """
 
-        self.feature -= entry
-        self.entries.remove(entry)
-    
-    
+
     def merge(self, node):
         """!
         @brief Merge leaf node to the current.
@@ -685,8 +627,8 @@ class leaf_node(cfnode):
         # Move entries from merged node
         for entry in node.entries:
             self.entries.append(entry)
-            
-    
+
+
     def get_farthest_entries(self, type_measurement):
         """!
         @brief Find pair of farthest entries of the node.
@@ -714,8 +656,8 @@ class leaf_node(cfnode):
                     farthest_entity2 = candidate2
         
         return [farthest_entity1, farthest_entity2]
-    
-    
+
+
     def get_nearest_index_entry(self, entry, type_measurement):
         """!
         @brief Find nearest index of nearest entry of node for the specified entry.
@@ -728,16 +670,17 @@ class leaf_node(cfnode):
         """
         
         minimum_distance = float('Inf')
-        nearest_index = 0
+        nearest_index = -1
         
         for candidate_index in range(0, len(self.entries)):
             candidate_distance = self.entries[candidate_index].get_distance(entry, type_measurement)
             if candidate_distance < minimum_distance:
+                minimum_distance = candidate_distance
                 nearest_index = candidate_index
         
         return nearest_index
-    
-    
+
+
     def get_nearest_entry(self, entry, type_measurement):
         """!
         @brief Find nearest entry of node for the specified entry.
@@ -750,7 +693,7 @@ class leaf_node(cfnode):
         """
         
         min_key = lambda cur_entity: cur_entity.get_distance(entry, type_measurement)
-        return min(self.__entries, key=min_key)
+        return min(self.entries, key=min_key)
 
 
 class cftree:
@@ -759,7 +702,7 @@ class cftree:
     @details A CF-tree is a height-balanced tree with two parameters: branching factor and threshold.
     
     """
-    
+
     @property
     def root(self):
         """!
@@ -767,17 +710,17 @@ class cftree:
         
         """
         return self.__root
-    
-    
+
+
     @property
     def leafes(self):
         """!
-        @return (list) List of all non-leaf nodes in the tree.
+        @return (list) List of all leaf nodes in the tree.
         
         """
         return self.__leafes
-    
-    
+
+
     @property
     def amount_nodes(self):
         """!
@@ -785,8 +728,8 @@ class cftree:
         
         """
         return self.__amount_nodes
-    
-    
+
+
     @property
     def amount_entries(self):
         """!
@@ -794,8 +737,8 @@ class cftree:
         
         """
         return self.__amount_entries
-    
-    
+
+
     @property
     def height(self):
         """!
@@ -803,8 +746,8 @@ class cftree:
         
         """
         return self.__height
-    
-    
+
+
     @property
     def branch_factor(self):
         """!
@@ -813,8 +756,8 @@ class cftree:
         
         """
         return self.__branch_factor
-    
-    
+
+
     @property
     def threshold(self):
         """!
@@ -822,8 +765,8 @@ class cftree:
         
         """
         return self.__threshold
-    
-    
+
+
     @property
     def max_entries(self):
         """!
@@ -831,8 +774,8 @@ class cftree:
         
         """
         return self.__max_entries
-    
-    
+
+
     @property
     def type_measurement(self):
         """!
@@ -840,8 +783,8 @@ class cftree:
         
         """
         return self.__type_measurement
-    
-    
+
+
     def __init__(self, branch_factor, max_entries, threshold, type_measurement = measurement_type.CENTROID_EUCLIDEAN_DISTANCE):
         """!
         @brief Create CF-tree.
@@ -870,8 +813,8 @@ class cftree:
         self.__amount_nodes = 0    # root, despite it can be None.
         self.__amount_entries = 0
         self.__height = 0          # tree size with root.
-    
-    
+
+
     def get_level_nodes(self, level):
         """!
         @brief Traverses CF-tree to obtain nodes at the specified level.
@@ -882,13 +825,13 @@ class cftree:
         
         """
         
-        level_nodes = [];
-        if (level < self.__height):
-            level_nodes = self.__recursive_get_level_nodes(level, self.__root);
+        level_nodes = []
+        if level < self.__height:
+            level_nodes = self.__recursive_get_level_nodes(level, self.__root)
         
-        return level_nodes;
-    
-    
+        return level_nodes
+
+
     def __recursive_get_level_nodes(self, level, node):
         """!
         @brief Traverses CF-tree to obtain nodes at the specified level recursively.
@@ -900,28 +843,27 @@ class cftree:
         
         """
         
-        level_nodes = [];
-        if (level is 0):
-            level_nodes.append(node);
+        level_nodes = []
+        if level is 0:
+            level_nodes.append(node)
         
         else:
             for sucessor in node.successors:
-                level_nodes += self.__recursive_get_level_nodes(level - 1, sucessor);
+                level_nodes += self.__recursive_get_level_nodes(level - 1, sucessor)
         
-        return level_nodes;
-    
-    
-    def insert_cluster(self, cluster):
+        return level_nodes
+
+
+    def insert_point(self, point):
         """!
-        @brief Insert cluster that is represented as list of points where each point is represented by list of coordinates.
-        @details Clustering feature is created for that cluster and inserted to the tree.
-        
-        @param[in] cluster (list): Cluster that is represented by list of points that should be inserted to the tree.
-        
+        @brief Insert point that is represented by list of coordinates.
+
+        @param[in] point (list): Point represented by list of coordinates that should be inserted to CF tree.
+
         """
-        
-        entry = cfentry(len(cluster), linear_sum(cluster), square_sum(cluster));
-        self.insert(entry);
+
+        entry = cfentry(len([point]), linear_sum([point]), square_sum([point]))
+        self.insert(entry)
     
     
     def insert(self, entry):
@@ -932,24 +874,24 @@ class cftree:
         
         """
                 
-        if (self.__root is None):
-            node = leaf_node(entry, None, [ entry ], None);
+        if self.__root is None:
+            node = leaf_node(entry, None, [entry])
             
-            self.__root = node;
-            self.__leafes.append(node);
+            self.__root = node
+            self.__leafes.append(node)
             
             # Update statistics
-            self.__amount_entries += 1;
-            self.__amount_nodes += 1;
-            self.__height += 1;             # root has successor now
+            self.__amount_entries += 1
+            self.__amount_nodes += 1
+            self.__height += 1             # root has successor now
         else:
-            child_node_updation = self.__recursive_insert(entry, self.__root);
-            if (child_node_updation is True):
+            child_node_updation = self.__recursive_insert(entry, self.__root)
+            if child_node_updation is True:
                 # Splitting has been finished, check for possibility to merge (at least we have already two children).
-                if (self.__merge_nearest_successors(self.__root) is True):
-                    self.__amount_nodes -= 1;
-    
-    
+                if self.__merge_nearest_successors(self.__root) is True:
+                    self.__amount_nodes -= 1
+
+
     def find_nearest_leaf(self, entry, search_node = None):
         """!
         @brief Search nearest leaf to the specified clustering feature.
@@ -961,20 +903,20 @@ class cftree:
         
         """
         
-        if (search_node is None):
-            search_node = self.__root;
+        if search_node is None:
+            search_node = self.__root
         
-        nearest_node = search_node;
+        nearest_node = search_node
         
-        if (search_node.type == cfnode_type.CFNODE_NONLEAF):
-            min_key = lambda child_node: child_node.feature.get_distance(entry, self.__type_measurement);
-            nearest_child_node = min(search_node.successors, key = min_key);
+        if search_node.type == cfnode_type.CFNODE_NONLEAF:
+            min_key = lambda child_node: child_node.feature.get_distance(entry, self.__type_measurement)
+            nearest_child_node = min(search_node.successors, key = min_key)
             
-            nearest_node = self.find_nearest_leaf(entry, nearest_child_node);
+            nearest_node = self.find_nearest_leaf(entry, nearest_child_node)
         
-        return nearest_node;
-    
-    
+        return nearest_node
+
+
     def __recursive_insert(self, entry, search_node):
         """!
         @brief Recursive insert of the entry to the tree.
@@ -988,14 +930,14 @@ class cftree:
         """
         
         # None-leaf node
-        if (search_node.type == cfnode_type.CFNODE_NONLEAF):
-            return self.__insert_for_noneleaf_node(entry, search_node);
+        if search_node.type == cfnode_type.CFNODE_NONLEAF:
+            return self.__insert_for_noneleaf_node(entry, search_node)
         
         # Leaf is reached 
         else:
-            return self.__insert_for_leaf_node(entry, search_node);
-    
-    
+            return self.__insert_for_leaf_node(entry, search_node)
+
+
     def __insert_for_leaf_node(self, entry, search_node):
         """!
         @brief Recursive insert entry from leaf node to the tree.
@@ -1007,32 +949,33 @@ class cftree:
         
         """
         
-        node_amount_updation = False;
+        node_amount_updation = False
         
         # Try to absorb by the entity
-        index_nearest_entry = search_node.get_nearest_index_entry(entry, self.__type_measurement);
-        merged_entry = search_node.entries[index_nearest_entry] + entry;
+        index_nearest_entry = search_node.get_nearest_index_entry(entry, self.__type_measurement)
+        nearest_entry = search_node.entries[index_nearest_entry]    # get nearest entry
+        merged_entry = nearest_entry + entry
         
         # Otherwise try to add new entry
-        if (merged_entry.get_diameter() > self.__threshold):
+        if merged_entry.get_diameter() > self.__threshold:
             # If it's not exceeded append entity and update feature of the leaf node.
-            search_node.insert_entry(entry);
+            search_node.insert_entry(entry)
             
             # Otherwise current node should be splitted
-            if (len(search_node.entries) > self.__max_entries):
-                self.__split_procedure(search_node);
-                node_amount_updation = True;
+            if len(search_node.entries) > self.__max_entries:
+                self.__split_procedure(search_node)
+                node_amount_updation = True
             
             # Update statistics
-            self.__amount_entries += 1;
+            self.__amount_entries += 1
             
         else:
-            search_node.entries[index_nearest_entry] = merged_entry;
-            search_node.feature += entry;
+            search_node.entries[index_nearest_entry] = merged_entry
+            search_node.feature += entry
         
-        return node_amount_updation;
-    
-    
+        return node_amount_updation
+
+
     def __insert_for_noneleaf_node(self, entry, search_node):
         """!
         @brief Recursive insert entry from none-leaf node to the tree.
@@ -1044,48 +987,48 @@ class cftree:
         
         """
         
-        node_amount_updation = False;
+        node_amount_updation = False
         
-        min_key = lambda child_node: child_node.get_distance(search_node, self.__type_measurement);
-        nearest_child_node = min(search_node.successors, key = min_key);
+        min_key = lambda child_node: child_node.get_distance(search_node, self.__type_measurement)
+        nearest_child_node = min(search_node.successors, key=min_key)
         
-        child_node_updation = self.__recursive_insert(entry, nearest_child_node);
+        child_node_updation = self.__recursive_insert(entry, nearest_child_node)
         
         # Update clustering feature of none-leaf node.
-        search_node.feature += entry;
+        search_node.feature += entry
             
         # Check branch factor, probably some leaf has been splitted and threshold has been exceeded.
         if (len(search_node.successors) > self.__branch_factor):
             
             # Check if it's aleady root then new root should be created (height is increased in this case).
-            if (search_node is self.__root):
-                self.__root = non_leaf_node(search_node.feature, None, [ search_node ], None);
-                search_node.parent = self.__root;
+            if search_node is self.__root:
+                self.__root = non_leaf_node(search_node.feature, None, [search_node])
+                search_node.parent = self.__root
                 
                 # Update statistics
-                self.__amount_nodes += 1;
-                self.__height += 1;
+                self.__amount_nodes += 1
+                self.__height += 1
                 
-            [new_node1, new_node2] = self.__split_nonleaf_node(search_node);
+            [new_node1, new_node2] = self.__split_nonleaf_node(search_node)
             
             # Update parent list of successors
-            parent = search_node.parent;
-            parent.successors.remove(search_node);
-            parent.successors.append(new_node1);
-            parent.successors.append(new_node2);
+            parent = search_node.parent
+            parent.successors.remove(search_node)
+            parent.successors.append(new_node1)
+            parent.successors.append(new_node2)
             
             # Update statistics
-            self.__amount_nodes += 1;
-            node_amount_updation = True;
+            self.__amount_nodes += 1
+            node_amount_updation = True
             
-        elif (child_node_updation is True):
+        elif child_node_updation is True:
             # Splitting has been finished, check for possibility to merge (at least we have already two children).
-            if (self.__merge_nearest_successors(search_node) is True):
-                self.__amount_nodes -= 1;
+            if self.__merge_nearest_successors(search_node) is True:
+                self.__amount_nodes -= 1
         
-        return node_amount_updation;
-    
-    
+        return node_amount_updation
+
+
     def __merge_nearest_successors(self, node):
         """!
         @brief Find nearest sucessors and merge them.
@@ -1096,23 +1039,23 @@ class cftree:
         
         """
         
-        merging_result = False;
+        merging_result = False
         
-        if (node.successors[0].type == cfnode_type.CFNODE_NONLEAF):
-            [nearest_child_node1, nearest_child_node2] = node.get_nearest_successors(self.__type_measurement);
+        if node.successors[0].type == cfnode_type.CFNODE_NONLEAF:
+            [nearest_child_node1, nearest_child_node2] = node.get_nearest_successors(self.__type_measurement)
             
-            if (len(nearest_child_node1.successors) + len(nearest_child_node2.successors) <= self.__branch_factor):
-                node.successors.remove(nearest_child_node2);
-                if (nearest_child_node2.type == cfnode_type.CFNODE_LEAF):
-                    self.__leafes.remove(nearest_child_node2);
+            if len(nearest_child_node1.successors) + len(nearest_child_node2.successors) <= self.__branch_factor:
+                node.successors.remove(nearest_child_node2)
+                if nearest_child_node2.type == cfnode_type.CFNODE_LEAF:
+                    self.__leafes.remove(nearest_child_node2)
                 
-                nearest_child_node1.merge(nearest_child_node2);
+                nearest_child_node1.merge(nearest_child_node2)
                 
-                merging_result = True;
+                merging_result = True
         
-        return merging_result;
-    
-    
+        return merging_result
+
+
     def __split_procedure(self, split_node):
         """!
         @brief Starts node splitting procedure in the CF-tree from the specify node.
@@ -1120,30 +1063,30 @@ class cftree:
         @param[in] split_node (cfnode): CF-tree node that should be splitted.
         
         """
-        if (split_node is self.__root):
-            self.__root = non_leaf_node(split_node.feature, None, [ split_node ], None);
-            split_node.parent = self.__root;
+        if split_node is self.__root:
+            self.__root = non_leaf_node(split_node.feature, None, [ split_node ])
+            split_node.parent = self.__root
             
             # Update statistics
-            self.__amount_nodes += 1;
-            self.__height += 1;
+            self.__amount_nodes += 1
+            self.__height += 1
         
-        [new_node1, new_node2] = self.__split_leaf_node(split_node);
+        [new_node1, new_node2] = self.__split_leaf_node(split_node)
         
-        self.__leafes.remove(split_node);
-        self.__leafes.append(new_node1);
-        self.__leafes.append(new_node2);
+        self.__leafes.remove(split_node)
+        self.__leafes.append(new_node1)
+        self.__leafes.append(new_node2)
         
         # Update parent list of successors
-        parent = split_node.parent;
-        parent.successors.remove(split_node);
-        parent.successors.append(new_node1);
-        parent.successors.append(new_node2);
+        parent = split_node.parent
+        parent.successors.remove(split_node)
+        parent.successors.append(new_node1)
+        parent.successors.append(new_node2)
         
         # Update statistics
-        self.__amount_nodes += 1;
-    
-    
+        self.__amount_nodes += 1
+
+
     def __split_nonleaf_node(self, node):
         """!
         @brief Performs splitting of the specified non-leaf node.
@@ -1154,29 +1097,29 @@ class cftree:
         
         """
         
-        [farthest_node1, farthest_node2] = node.get_farthest_successors(self.__type_measurement);
+        [farthest_node1, farthest_node2] = node.get_farthest_successors(self.__type_measurement)
         
         # create new non-leaf nodes
-        new_node1 = non_leaf_node(farthest_node1.feature, node.parent, [ farthest_node1 ], None);
-        new_node2 = non_leaf_node(farthest_node2.feature, node.parent, [ farthest_node2 ], None);
+        new_node1 = non_leaf_node(farthest_node1.feature, node.parent, [farthest_node1])
+        new_node2 = non_leaf_node(farthest_node2.feature, node.parent, [farthest_node2])
         
-        farthest_node1.parent = new_node1;
-        farthest_node2.parent = new_node2;
+        farthest_node1.parent = new_node1
+        farthest_node2.parent = new_node2
         
         # re-insert other successors
         for successor in node.successors:
-            if ( (successor is not farthest_node1) and (successor is not farthest_node2) ):
-                distance1 = new_node1.get_distance(successor, self.__type_measurement);
-                distance2 = new_node2.get_distance(successor, self.__type_measurement);
+            if (successor is not farthest_node1) and (successor is not farthest_node2):
+                distance1 = new_node1.get_distance(successor, self.__type_measurement)
+                distance2 = new_node2.get_distance(successor, self.__type_measurement)
                 
-                if (distance1 < distance2):
-                    new_node1.insert_successor(successor);
+                if distance1 < distance2:
+                    new_node1.insert_successor(successor)
                 else:
-                    new_node2.insert_successor(successor);
+                    new_node2.insert_successor(successor)
         
-        return [new_node1, new_node2];
-    
-    
+        return [new_node1, new_node2]
+
+
     def __split_leaf_node(self, node):
         """!
         @brief Performs splitting of the specified leaf node.
@@ -1190,45 +1133,21 @@ class cftree:
         """
         
         # search farthest pair of entries
-        [farthest_entity1, farthest_entity2] = node.get_farthest_entries(self.__type_measurement);
+        [farthest_entity1, farthest_entity2] = node.get_farthest_entries(self.__type_measurement)
                     
         # create new nodes
-        new_node1 = leaf_node(farthest_entity1, node.parent, [ farthest_entity1 ], None);
-        new_node2 = leaf_node(farthest_entity2, node.parent, [ farthest_entity2 ], None);
+        new_node1 = leaf_node(farthest_entity1, node.parent, [farthest_entity1])
+        new_node2 = leaf_node(farthest_entity2, node.parent, [farthest_entity2])
         
         # re-insert other entries
         for entity in node.entries:
-            if ( (entity is not farthest_entity1) and (entity is not farthest_entity2) ):
-                distance1 = new_node1.feature.get_distance(entity, self.__type_measurement);
-                distance2 = new_node2.feature.get_distance(entity, self.__type_measurement);
+            if (entity is not farthest_entity1) and (entity is not farthest_entity2):
+                distance1 = new_node1.feature.get_distance(entity, self.__type_measurement)
+                distance2 = new_node2.feature.get_distance(entity, self.__type_measurement)
                 
-                if (distance1 < distance2):
-                    new_node1.insert_entry(entity);
+                if distance1 < distance2:
+                    new_node1.insert_entry(entity)
                 else:
-                    new_node2.insert_entry(entity);
+                    new_node2.insert_entry(entity)
         
-        return [new_node1, new_node2];
-    
-    
-    def show_feature_destibution(self, data = None):
-        """!
-        @brief Shows feature distribution.
-        @details Only features in 1D, 2D, 3D space can be visualized.
-        
-        @param[in] data (list): List of points that will be used for visualization, if it not specified than feature will be displayed only.
-        
-        """
-        visualizer = cluster_visualizer();
-        
-        print("amount of nodes: ", self.__amount_nodes);
-        
-        if (data is not None):
-            visualizer.append_cluster(data, marker = 'x');
-        
-        for level in range(0, self.height):
-            level_nodes = self.get_level_nodes(level);
-            
-            centers = [ node.feature.get_centroid() for node in level_nodes ];
-            visualizer.append_cluster(centers, None, markersize = (self.height - level + 1) * 5);
-        
-        visualizer.show();
\ No newline at end of file
+        return [new_node1, new_node2]
